-module(kinds).

-public_type core_node() = cerl:core_node().
-public_type clause() = cerl:clause().
-abstype constructor() = {atomic, any()} + cons + {tuple, int()} + vector.

-public_type fdef() = core_node().

-type list(X) = [X | list(X)] + [].

- type char() = int().
- type string() = list(char()).

type(core_node()) ->
    exp + mdef + fdef + clause.

exp_type(core_node()) ->
    var + alias + constructor + control + dynamic + fref + local_call + apply + call + 'fun'.

constructor_exp_arg_list(core_node()) ->
    list(core_node()).

exp_constructor(core_node()) ->
    constructor().

constructor_type(constructor()) ->
    atomic + cons + tuple + vector.

concrete(core_node()) ->
    any().    

%%% Literals

is_literal(core_node()) ->
    true + false.

is_string(core_node()) ->
    true + false.

is_int(core_node()) ->
    true + false.

is_atom(core_node()) ->
    true + false.

atom_val({atomic, _}) ->
    atom().

char_val({atomic, _}) ->
    char().

int_val({atomic, _}) ->
    int().

float_val({atomic, _}) ->
    float().

string_val({atomic, _}) ->
    string().


control_exp_type(core_node()) ->
    'case' + 'let' + letrec + 'receive' + 'catch' + 'try' + 'seq'.    

%%% Let

let_exp_var_list(core_node()) ->
    list(cerl:variable()).

let_exp_arg(core_node()) ->
    core_node().

let_exp_body(core_node()) ->
    core_node().

%%% Letrec

letrec_def_list(core_node()) ->
    list(core_node()).

letrec_body(core_node()) ->
    core_node().

%%% Case
case_exp_arg(core_node()) ->
    core_node().

case_exp_clause_list(core_node()) ->
    list(clause()).

clause_guard(clause()) ->
    core_node().

clause_body(clause()) ->
    core_node().

clause_pattern_list(clause()) ->
    list(core_node()).

%%% Receive

receive_exp_action(core_node()) ->
    core_node().
receive_exp_clause_list(core_node()) ->
    list(clause()).
receive_exp_timeout(core_node()) ->
    core_node().

seq_arg(core_node()) ->
    core_node().

seq_body(core_node()) ->
    core_node().

catch_body(core_node()) ->
    core_node().

try_body(core_node()) ->
    core_node().

%%% Fun

fun_var_list(core_node()) ->
    list(core_node()).

fun_body(core_node()) ->
    core_node().

%%% Fref

is_fref(core_node()) ->
    true + false.

fref_name(core_node()) ->
    atom().

fref_arity(core_node()) ->
    int().
	

%%% Apply

apply_op(core_node()) ->
    core_node().

apply_arg_list(core_node()) ->
    list(core_node()).

%%% Call

call_module(core_node()) ->
    core_node().

call_name(core_node()) ->
    core_node().

call_arg_list(core_node()) ->
    list(core_node()).

%%% Variable

var_name(core_node()) ->
    any().

%%% Fdef

is_fdef(core_node()) ->
    true + false.

fdef_arity(core_node()) ->
    int().

fdef_body(core_node()) ->
    core_node().

%%% MDef

mdef_export_list(core_node()) ->
    list(any()).

mdef_body(core_node()) ->
    core_node().

%%% Annotations

add_ann(core_node(), any()) ->
    core_node().


%%% Pattern

pattern_type(core_node()) ->
    var + alias + constructor.

alias_var(core_node()) ->
    cerl:variable().

alias_pattern(core_node()) ->
    core_node().

constructor_pattern_arg_list(core_node()) ->
    list(core_node()).

pattern_constructor(core_node()) ->
    {atomic, core_node()} + cons + {tuple, int()}.

atomic_constructor_type({atomic, _}) ->
    int + float + nil + string.


%%% Guard

guard_type(core_node()) ->
    true + comparision + control.


%%%

pos(core_node()) ->
    int().

name_atom(core_node()) ->
    atom().

%%%

read_module(file:descr(), atom()) ->
    core_node().

read_module(file:descr(), atom(), any()) ->
    core_node().
