%%% File    : env.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : Environments
%%% Created : 22 Apr 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

-module(env).
-author('svenolof@harpo.csd.uu.se').

-export([empty/0, bind/3, bind_list/3, lookup/2, defined/2]).

bind(Name, Value, Env) ->
    case gb_trees:lookup(Name, Env) of
	none ->
	    gb_trees:insert(Name, Value, Env);
	{value, _} ->
	    gb_trees:enter(Name, Value, Env)
    end.

bind_list([Name | NameList], [Value | ValueList], Env) ->
    Env1 = bind(Name, Value, Env),
    bind_list(NameList, ValueList, Env1);
bind_list([], [], Env) ->
    Env.

lookup(Name, Env) ->
    {value, Value} = gb_trees:lookup(Name, Env),
    Value.

defined(Name, Env) ->
    gb_trees:is_defined(Name, Env).

empty() ->
    gb_trees:empty().



