%%% File    : kinds.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : 
%%% Created :  7 Apr 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>


-module(kinds).
-author('svenolof@harpo.csd.uu.se').

% -compile(export_all).

-export([type/1, exp_type/1, exp_constructor/1,
	 constructor_exp_arg_list/1, make_constructor_exp/2,
	 is_literal/1, concrete/1, is_string/1,
	 constructor_type/1, atomic_constructor_type/1, atom_val/1,
	 int_val/1, is_int/1, float_val/1, char_val/1, string_val/1,
	 make_atom/1, is_atom/1,
	 control_exp_type/1, let_exp_var_list/1, let_exp_arg/1,
	 let_exp_body/1, make_let/3, 
	 letrec_def_list/1, letrec_body/1, make_letrec/2,
	 case_exp_arg/1,
	 case_exp_clause_list/1, make_case/2, clause_pattern_list/1,
	 clause_guard/1, clause_body/1, make_clause/3,
	 receive_exp_clause_list/1, receive_exp_timeout/1,
	 receive_exp_action/1, make_receive/3, make_receive/1,
	 seq_arg/1, seq_body/1, make_seq/2,
	 catch_body/1, make_catch/1,
	 try_arg/1, try_var_list/1, try_body/1, 
	 try_e_var_list/1, try_handler/1,
	 make_try/5,
	 pattern_type/1, pattern_constructor/1,
	 constructor_pattern_arg_list/1, make_constructor_pattern/2,
	 alias_var/1, alias_pattern/1, make_alias/2,
	 var_name/1, make_var/1, fun_var_list/1, fun_arity/1, fun_body/1,
	 make_fun/2, make_fref/2, fref_name/1, fref_arity/1, is_fref/1, 
         call_module/1,
	 call_name/1, call_arg_list/1, make_call/3, call_type/1,
	 call_mfa/1, dynamic_mod/1, dynamic_name/1, dynamic_arity/1,
	 make_dynamic/3, guard_type/1, comparision_arg_list/1,
	 comparision_guard_op/1, fdef_name/1,
	 fdef_arity/1, is_fdef/1, 
	 apply_op/1, apply_arg_list/1, make_apply/2,
	 fdef_body/1, make_fdef/3, mdef_name/1,
	 mdef_body/1, mdef_export_list/1, make_mdef/3, copy_ann/2,
	 add_ann/2, read_module/2, read_module/3, pos/1, name_atom/1]).


c_type(Node) ->
    case is_fdef(Node) of
	true ->
	    fdef;
	false ->
	    cerl:type(Node)
    end.


type(Node) ->
    Type = c_type(Node),
    case Type of
	var -> exp;
	alias -> exp; % Strange...
	atom -> exp;
	char -> exp;
	cons -> exp;
	float -> exp;
	int -> exp;
	local -> exp;
	nil -> exp;
	string -> exp;
	tuple -> exp;
	values -> exp;
	'case' -> exp;    
	'let' -> exp;    
	letrec -> exp;
	'receive' -> exp;    
	'catch' -> exp;
	'try' -> exp;
	seq -> exp;    
	dynamic -> exp;
	fref -> exp;
	apply -> exp;
	call -> exp;
	primop -> exp; % Buu
	'fun' -> exp;
	module -> mdef;
	fdef -> fdef;
	clause -> clause;
	literal -> exp
    end.
    


exp_type(Exp) ->
    case c_type(Exp) of
	var -> var;
	alias -> alias;
	atom -> constructor;
	char -> constructor;
	cons -> constructor;
	float -> constructor;
	int -> constructor;
	local -> constructor;
	nil -> constructor;
	tuple -> constructor;
	values -> 
	    constructor;
	literal -> constructor;
	'case' -> control;    
	'let' -> control;    
	letrec -> control;
	'receive' -> control;    
	seq -> control;    
	'catch' -> control;
	'try' -> control;    
	dynamic -> dynamic;
	fref -> fref;
	apply -> 
	    case cerl:type(cerl:apply_op(Exp)) of
		fref ->
		    local_call;
		_ ->
		    apply
	    end;		
	call -> call;
	primop -> call;
	'fun' -> 'fun'
    end.     

exp_constructor(Exp) ->
    case cerl:type(Exp) of
	literal ->
	    C = cerl:concrete(Exp),
	    if 
		is_atom(C) -> 
		    {atomic, Exp};
		is_integer(C)  ->
		    {atomic, Exp};
		is_float(C) ->
		    {atomic, Exp};
		C == [] ->
		    {atomic, Exp};
		is_list(C) ->
		    cons; % We have already tested for nil!
		is_tuple(C) ->
		    {tuple, cerl:data_arity(Exp)}
	    end;
	cons ->
	    cons;
	tuple ->
	    {tuple, cerl:data_arity(Exp)};
	values ->
	     case cerl:values_es(Exp) of
		[Exp1] ->
		    exp_constructor(Exp1);
		_ ->
		    vector
	    end
    end.

is_literal(Exp) ->
    c_type(Exp) == literal.

    
is_string(_Exp) ->
    false.

%% is_cons(Exp) ->
%%     cerl:is_c_cons(Exp).

constructor_exp_arg_list(Exp) ->
    case cerl:type(Exp) of
	values ->
	    cerl:values_es(Exp);
	_ ->
	    cerl:data_es(Exp)
   end.
  


make_constructor_exp({atomic, Exp}, []) ->
    Exp;
make_constructor_exp({tuple, _}, ExpList) ->
    cerl:c_tuple(ExpList);
make_constructor_exp(vector, ExpList) ->
    cerl:c_values(ExpList);
make_constructor_exp(cons, [Head, Tail]) ->
    cerl:c_cons(Head, Tail).

constructor_type({atomic, _}) ->
    atomic;
constructor_type({tuple, _}) ->
    tuple;
constructor_type(vector) ->
    vector;
constructor_type(C) ->
    C.

atomic_constructor_type({atomic, Node}) ->
    C = cerl:concrete(Node),
    if integer(C) ->
	    int;
       float(C) ->
	    float;
       atom(C) ->
	    atom;
       C ==[] ->
	    nil;
       list(C) ->
	    string
    end.

concrete(E) ->
    cerl:concrete(E).

int_val({atomic,A}) ->
    cerl:int_val(A);
int_val(Node) ->
    cerl:int_val(Node).

is_int(Exp) ->
    cerl:is_c_int(Exp).

%% is_nil(Exp) ->
%%     cerl:is_c_nil(Exp).



float_val({atomic,A}) ->
    cerl:float_val(A).

atom_val({atomic,A}) ->
    list_to_atom(cerl:atom_name(A)).

char_val({atomic,A}) ->
    cerl:char_val(A).

string_val({atomic,A}) ->
    cerl:concrete(A).

make_atom(Atom) ->
    cerl:c_atom(Atom).

is_atom(Exp) ->
    cerl:is_c_atom(Exp).

%% Control
	
control_exp_type(Exp) ->
    case c_type(Exp) of
	'case' -> 'case';    
	'let' -> 'let';    
	letrec -> letrec;
	'receive' -> 'receive';    
	'catch' -> 'catch';
	seq -> 	seq;    
	'try' -> 'try'
    end.     

%% Let

let_exp_var_list(Exp) ->
    cerl:let_vars(Exp).

let_exp_arg(Exp) ->
    cerl:let_arg(Exp).

let_exp_body(Exp) ->
    cerl:let_body(Exp).

make_let(Var, Arg, Exp) ->
    cerl:c_let(Var, Arg, Exp).

%% Letrec

letrec_def_list(Exp) ->
    cerl:letrec_defs(Exp).

letrec_body(Exp) ->
    cerl:letrec_body(Exp).

make_letrec(DefList, Body) ->
    cerl:c_letrec(DefList, Body).



%% Case

case_exp_arg(Exp) ->
    cerl:case_arg(Exp).

case_exp_clause_list(Exp) ->
    cerl:case_clauses(Exp).

make_case(Arg, ClauseList) ->
    cerl:c_case(Arg, ClauseList).

%% Clause

clause_pattern_list(Cl) ->
    cerl:clause_pats(Cl).

clause_guard(Cl) ->
    cerl:clause_guard(Cl).

clause_body(Cl) ->
    cerl:clause_body(Cl).

make_clause(Pattern, Guard, Body) ->
    cerl:c_clause(Pattern, Guard, Body).

%%% Receive

receive_exp_clause_list(Exp) ->
    cerl:receive_clauses(Exp).

receive_exp_timeout(Exp) ->
    cerl:receive_timeout(Exp).

receive_exp_action(Exp) ->
    cerl:receive_action(Exp).

make_receive(ClauseList) ->
    cerl:c_receive(ClauseList).

make_receive(ClauseList, Timeout, Action) ->
    cerl:c_receive(ClauseList, Timeout, Action).

%%% Seq

seq_arg(Exp) ->
    cerl:seq_arg(Exp).

seq_body(Exp) ->
    cerl:seq_body(Exp).

make_seq(Arg, Body) ->
    cerl:c_seq(Arg, Body).

%%% Catch

catch_body(Exp) ->
    cerl:catch_body(Exp).

make_catch(Body) ->
    cerl:c_catch(Body).


%%% Try

try_arg(Exp) ->
    cerl:try_arg(Exp).

try_var_list(Exp) ->
    cerl:try_vars(Exp).

try_body(Exp) ->
    cerl:try_body(Exp).

try_e_var_list(Exp) ->
    cerl:try_evars(Exp).

try_handler(Exp) ->
    cerl:try_handler(Exp).


make_try(Exp, VarList, Body, Evars, Handler) ->
    cerl:c_try(Exp, VarList, Body, Evars, Handler).

%%% Pattern

pattern_type(Pattern) ->
    case c_type(Pattern) of
	var -> var;
	alias -> alias;
	atom -> constructor;
	char -> constructor;
	cons -> constructor;
	float -> constructor;
	int -> constructor;
	local -> constructor;
	nil -> constructor;
	string -> constructor;
	tuple -> constructor;
	literal -> constructor
    end.     

alias_var(Pattern) ->
    cerl:alias_var(Pattern).

alias_pattern(Pattern) ->
    cerl:alias_pat(Pattern).

make_alias(Var, Pattern) ->
    cerl:c_alias(Var, Pattern).


pattern_constructor(Pattern) ->
    case cerl:data_type(Pattern) of
	{atomic, _} ->
	    {atomic, Pattern};
	cons ->
	    cons;
	tuple ->
	    {tuple, cerl:data_arity(Pattern)}
    end.

constructor_pattern_arg_list(Pattern) ->
    case cerl:data_type(Pattern) of
	{string, _} ->
	   case cerl:string_val(Pattern) of
	       [Char | RestString] ->
		   Car = cerl:c_int(Char),
		   Cdr = cerl:c_string(RestString),
		   [Car, Cdr];
	       [] ->
		   []
	   end;
	_ ->
	   cerl:data_es(Pattern)
   end.

       

make_constructor_pattern({atomic, Pattern}, []) ->
    Pattern;
make_constructor_pattern({tuple, _}, PatternList) ->
    cerl:c_tuple(PatternList);
make_constructor_pattern(cons, [Head, Tail]) ->
    cerl:c_cons(Head, Tail).


%%% Variable

var_name({var, _, Name}) ->
    Name;
var_name(Var) ->
    cerl:var_name(Var).

make_var(Name) ->
    cerl:c_var(Name).

%%% Fun

fun_var_list(Exp) ->
    cerl:fun_vars(Exp).

fun_arity(Exp) ->
    length(cerl:fun_vars(Exp)).    

fun_body(Exp) ->
    cerl:fun_body(Exp).

make_fun(VarList, Body) ->
    cerl:c_fun(VarList, Body).

%%% Fref

is_fref(Exp) ->
    case cerl:is_c_var(Exp) of
	true ->
	    case cerl:var_name(Exp) of
		{_,_} ->
		    true;
		_ -> 
		    false
	    end;
	false ->
	    false
    end.
    

fref_name(Exp) ->
    {Name, Arity} = cerl:var_name(Exp),
    Name.    

fref_arity(Exp) ->
    {Name, Arity} = cerl:var_name(Exp),
    Arity.

make_fref(Functor, Arity) ->
    cerl:c_var({Functor, Arity}).

%%% Apply

apply_op(Exp) ->
    cerl:apply_op(Exp).

apply_arg_list(Exp) ->
    cerl:apply_args(Exp).

make_apply(Op, ArgList) ->
    cerl:c_apply(Op, ArgList).

%%% Call 

call_module(Exp) ->
    case c_type(Exp)  of
	primop ->
	    make_atom('This is not the primop module');
	call ->
	    cerl:call_module(Exp)
    end.

call_name(Exp) ->
    case c_type(Exp)  of
	primop ->
	    cerl:primop_name(Exp);
	call ->
	    cerl:call_name(Exp)
    end.

call_arg_list(Exp) ->
    case c_type(Exp) of
	primop ->
	    cerl:primop_args(Exp);
	call ->
	    cerl:call_args(Exp)
    
    end.

make_call(Module, Name, ArgList) ->
    cerl:c_call(Module, Name, ArgList).
    
call_type(Exp) ->
    Module = call_module(Exp),
    Name = call_name(Exp),
    ArgList = call_arg_list(Exp),
    case kinds:is_atom(Module) and kinds:is_atom(Name) of
	true ->
	    case lists:member(cerl:atom_val(Module), 
			      [term, number, process, erlang]) of
		true ->
		    standard;
		false ->
		    general
	    end;
	false ->
	    general
    end.

call_mfa(Exp) ->
    {call_module(Exp),
     call_name(Exp),
     length(call_arg_list(Exp))}.


%%% Dynamics

dynamic_mod(Exp) ->
    cerl:dynamic_mod(Exp).

dynamic_name(Exp) ->
    cerl:dynamic_name(Exp).

dynamic_arity(Exp) ->
    cerl:dynamic_arity(Exp).

make_dynamic(Mod, Name, Arity) ->
    cerl:c_dynamic(Mod, Name, Arity).

%%% Guards

guard_type(Guard) ->
    case exp_type(Guard) of
	constructor ->
	    case constructor_type(exp_constructor(Guard)) of
		atomic ->
		    case atom_val(exp_constructor(Guard)) of
			true ->
			    true
		    end
	    end;
	call ->
	    case call_type(Guard) of
		standard ->
		    comparision
	    end;
	apply ->
	    comparision; % guard test list(..) is translated to Core
			 % Erlang code apply (fref list/1,..);
	control ->
	    control % whatever. This was added due to a bug in the 
                    % erl_to_core translation.
    end.

comparision_arg_list(Guard) ->
    call_arg_list(Guard).

comparision_guard_op(Guard) ->
    case exp_type(Guard) of
	call ->
	    case call_type(Guard) of
		standard ->
		    case call_mfa(Guard) of
			{term, '==', 2} ->
			    '==';
			{term, '=<', 2} ->
			    '=<'
		    end
	    end
    end.


%%% Fdef

fdef_name({Fref,Fun}) ->
    fref_name(Fref).
    

fdef_arity({Fref, Fun}) ->
    fref_arity(Fref).


fdef_body({Fref, Fun}) ->
    Fun;
fdef_body({Fref, Fun}) ->
    Fun.


make_fdef(Functor, Arity, Body) ->
    {make_fref(Functor, Arity), Body}.


is_fdef({Fref, Fun}) when tuple(Fref), tuple(Fun) ->
    (var == cerl:type(Fref)) 
	and ('fun' == cerl:type(Fun));
is_fdef(_) ->
    false.



%%%

mdef_name(Module) ->
    cerl:atom_name(cerl:module_name(Module)).
    

mdef_body(Module) ->
    cerl:module_defs(Module).

mdef_export_list(MDef) ->
    cerl:module_exports(MDef).

make_mdef(Name, ExportList, Body) ->
    cerl:c_module(cerl:c_atom(Name), ExportList, Body).

%%% Annotations

copy_ann(E1, E2) ->
    A = get_ann(E2),
    set_ann(E1, A).
    
add_ann(Node, Ann) ->
    A = get_ann(Node),
    set_ann(Node, [Ann | A]).

    
set_ann(Node, Ann) ->
    case is_fdef(Node) of
	false ->
	    cerl:set_ann(Node, Ann);
	true ->
	    {Fref, Fun} = Node,
	    Fref1 = cerl:set_ann(Fref, Ann),
	    {Fref1, Fun}
    end.

get_ann(Node) ->
    case is_fdef(Node) of
	false ->
	    cerl:get_ann(Node);
	true ->
	    {Fref, Fun} = Node,
	    cerl:get_ann(Fref)
    end.


%%% IO


read_module(FileName, Module) ->
    case compile:file(FileName,[dcore]) of
	{ok, _} ->
	    true;
	CompileResult ->
	    exit({bad_erlang_module, CompileResult, FileName})
    end,	    
    case sa_file:find_module(["."], Module, core) of
	{ok, CoreFileName} ->
	    {ok, B} = file:read_file(CoreFileName),
	    {ok, Ts, _} = cerl_scanner:string(binary_to_list(B)),
	    {ok, C} = cerl_parser:parse(Ts),
	    C;
	not_found ->
	    exit("Can't find Core Erlang file")
    end.

read_module(FileName, Module, Include) ->
    case compile:file(FileName,[{i,Include}, dcore]) of
	{ok, _} ->
	    true;
	CompileResult ->
	    exit({bad_erlang_module, CompileResult, FileName})
    end,	    
    case sa_file:find_module(["."], Module, core) of
	{ok, CoreFileName} ->
	    {ok, B} = file:read_file(CoreFileName),
% Richard & Robert likes to rename files.
	    {ok, Ts, _} = cerl_scanner:string(binary_to_list(B)),
%	    {ok, Ts, _} = core_scan:string(binary_to_list(B)),
	    {ok, C} = cerl_parser:parse(Ts),
%	    {ok, C} = core_parse:parse(Ts),
	    C;
	not_found ->
	    exit("Can't find Core Erlang file")
    end.

%%%

pos(Exp) ->
    L = get_ann(Exp),
    {value, {number, N}} = 
	lists:keysearch(number, 1, L),
    N.

name_atom(Node) ->
    case c_type(Node) of
	var ->
	    cerl:var_name(Node);
	module ->
	    CerlAtom = cerl:module_name(Node),
	    cerl:atom_val(CerlAtom);
	fdef ->
	    fdef_name(Node)
    end.

%    cerl:var_name(Node).


























