%%%----------------------------------------------------------------------
%%% File    : param.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : 
%%% Created : 18 Oct 2000 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%%----------------------------------------------------------------------

-module(param).
-author('svenolof@harpo.it.uu.se').

-export([initialize_environment/2, insert_arg_list/3, insert_fun_arg_list/4]).
-export([invent_arg_list/2,invent_fun_arg_list/3]).


initialize_environment(VarList, Context) ->
    Env = initialize_arg_list(VarList, Context),
    Self = tag:self(Context),
    Except = tag:except(Context),
    Env1 = env:bind_list(['$self','$except'], [Self, Except], Env),
    Env1.

initialize_arg_list(VarList, Context) ->
    initialize_arg_list(VarList, 1, env:empty(), Context).

initialize_arg_list([Var|VarList], Index, Env, Context) ->
    Term = tag:arg(Index, Context),
    Env1 = env:bind(Var, Term, Env) ,
    initialize_arg_list(VarList, Index+1, Env1, Context);
initialize_arg_list([], _, Env, _) ->
    Env.


insert_arg_list(AP, ArgList, AS) ->
    insert_arg_list(AP, ArgList, 1, AS).

insert_arg_list(AP, [Arg|ArgList], N, AS) ->
    AS1 = as:insert(tag:arg(N, AP), Arg, AS),
    insert_arg_list(AP, ArgList, N+1, AS1);
insert_arg_list(_, [], _, AS) ->
    AS.

insert_fun_arg_list(AP, Pos, ArgList, AS) ->
    insert_fun_arg_list(AP, Pos, ArgList, 1, AS).

insert_fun_arg_list(AP, Pos, [Arg|ArgList], N, AS) ->
    AS1 = as:insert(tag:fun_arg(AP, Pos, N), Arg, AS),
    insert_fun_arg_list(AP, Pos, ArgList, N+1, AS1);
insert_fun_arg_list(_, _,[], _, AS) ->
    AS.


% Previously in tag

invent_arg_list(N, Context) ->
    invent_arg_list_1(1, N, Context).

invent_arg_list_1(K, N, Context) when K =< N ->
    [tag:arg(K, Context) | 
     invent_arg_list_1(K+1, N, Context)];
invent_arg_list_1(K, N, _) when K > N ->
    [].


invent_fun_arg_list(N, Context, Pos) ->
    invent_fun_arg_list_1(1, N, Context, Pos).

invent_fun_arg_list_1(K, N, Context, Pos) when K =< N ->
    [tag:fun_arg(Context, Pos, K) | 
     invent_fun_arg_list_1(K+1, N, Context, Pos)];
invent_fun_arg_list_1(K, N, _, _) when K > N ->
    [].





