%% File    : spec_parse.yrl
%% Author  : Sven-Olof Nystrom (adapted from erl_parse)
%% Purpose : Syntax of warning files.

Nonterminals
form
attribute attr_val
utype_list utype_tuple utypes utype ptypes ptype funtype 
function 
expr 
exprs
atomic
points
point.

Terminals
atom float integer string var

'(' ')' ',' '->' 
'{' '}' '[' ']' '|' 
'end' 'fun' 
'/'
'-'
dot.

Rootsymbol form.


form -> attribute dot : '$1'.
form -> function dot : '$1'.

attribute -> '-' atom '(' attr_val ')' : build_attribute('$2', '$4').

attr_val -> exprs : '$1'.

utype_list -> '(' ')' : [].
utype_list -> '(' utypes ')' : '$2'.

utype_tuple -> '{' '}' : [].
utype_tuple -> '{' utypes '}' : '$2'.

utypes -> utype : ['$1'].
utypes -> utype ',' utypes : ['$1'|'$3'].

utype -> ptypes : {utype,'$1'}.


ptypes -> ptype : ['$1'].
ptypes -> ptypes '|' ptype : '$1' ++ ['$3'].

ptype -> var : element(3, '$1').
ptype -> '(' utype ')' : '$2'.
ptype -> '(' ')' : {utype,[]}.
ptype -> atom : {atom, element(2, '$1'), element(3, '$1')}.
ptype -> utype_tuple : {tuple,'$1'}.
ptype -> '[' ptype '|' ptype ']' : {cons, element(2, '$1'), '$2', '$4'}.
ptype -> '[' ']' : {nil, element(2, '$1')}.
ptype -> integer : '$1'.
ptype -> funtype : '$1'.

    
funtype ->
    'fun' utype_list '->' utype 'end' : {'fun', '$2', '$4'}.

function -> atom '/' integer '->' points :
	{funspec, element(2, '$2'), element(3, '$1'), element(3, '$3'), '$5'}.

points -> point ',' points : ['$1' | '$3'].
points -> point : ['$1'].

point -> atom ptype : {point, element(2, '$1'), element(3, '$1'), '$2'}.

expr -> var : '$1'.
expr -> atomic : '$1'.

% argument_list -> '(' ')' : {[],element(2, '$1')}.
% argument_list -> '(' exprs ')' : {'$2',element(2, '$1')}.

exprs -> expr : ['$1'].
exprs -> expr ',' exprs : ['$1' | '$3'].

atomic -> integer : '$1'.
atomic -> float : '$1'.
atomic -> atom : '$1'.
atomic -> string : '$1'.


Erlang code.

%% Some junk left from some other parser. Comments below make no sense
%% to me. / S-O

-author('rv@cslab.ericsson.se').
-copyright('Copyright (c) 1996 Ericsson Telecommunications AB').

%% This is a hack to help the JAM get better code.
-ifdef(JAM).
-compile([{parse_transform,jam_yecc_pj},pj]).
-endif.

-export([parse_form/1]).


%% Entry points compatible to old erl_parse.
%% These really suck and are only here until Calle gets multiple
%% entry points working.

parse_form(Tokens) ->
    parse(Tokens).

build_attribute({atom,La,module}, Val) ->
    case Val of
	[{atom,_,Module}] -> {attribute,La,module,Module};
	_ -> return_error(La, "bad module declaration")
    end;
build_attribute({atom,La,file}, Val) ->
    case Val of
	[{string,_,Name},{integer,_,Line}] ->
	    {attribute,La,file,{Name,Line}};
        _  -> return_error(La, "bad file declaration")
    end.
