%%% File    : present.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : Presentation interface
%%% Created : 18 May 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

-module(present).
-author('svenolof@harpo.csd.uu.se').

-export([present/2, present/3, present_function/2, sketch/2, generate_internal/1]).

-define(max_present, 999).    


present(Module, AS) -> 
    present(Module, standard_io, AS).


present(Module, Out, AS) -> 
    H = sa_file:open_output_file(Module, acc),
    io:format(H, "%    -*-  Erlang -*- ~n", []),
    Interesting = fun (Name) ->
			  tag:interesting(Name)
		  end,
    ContextTable = as:make_context_table(AS, Interesting),
    ContextList  = gb_trees:to_list(ContextTable),
    F = fun({Fnid, {ContextList2, NameList}}) ->
		present_function(Fnid, Out, H, ContextList2, NameList, AS)
	end,
    lists:foreach(F,  ContextList),
    file:close(H),
    true.

present_function(Fnid, AS) ->
    Module = program:fnid_module(Fnid),
    Out = standard_io,
    ACC = sa_file:open_output_file(Module, acc),
    Interesting = fun (Name) ->
			  tag:interesting(Name)
		  end,
    ContextTable = as:make_context_table(AS, Interesting),
    {value, {ContextList, NameList}} =  gb_trees:lookup(Fnid, ContextTable),
    present_function(Fnid, Out, ACC, ContextList, NameList, AS).

present_function(Fnid, Out,ACC,  ContextList, NameList, AS) ->
    NameList1 = lists:sort(NameList),
    RootList = make_root_list(ContextList, NameList1, AS),
    Struct = present_struct:collect_roots(NameList1, RootList, AS),
    case program:deid_type(Fnid) of 
	fnid ->
	    io:format(Out, "~n~nFunction ~s:~n", [program:show_fnid(Fnid)]),
	    MFA = {program:fnid_module(Fnid),
		   program:fnid_name(Fnid),
		   program:fnid_arity(Fnid)},
	    Struct2 = present_struct:rename(simplify(Struct)),
	    io:format("~n",[]),
	    show_struct(Out, Struct2),
	    {AList, WList} = draw:draw_struct(Struct2),
	    io:format(ACC, "~n~p.~n",[{function, MFA, AList, WList}]);
	tyid ->
	    true
    end.



make_root_list(ContextList, NameList, AS) ->
    F = fun (Name) ->
		G = fun (Context) ->
			    IP = term:make_ip(Name, Context),
			    Term = as:lookup(IP, AS),
			    Term
		    end,
		TermList = lists:map(G, ContextList),
		term:join(TermList)
	end,
    lists:map(F, NameList).



simplify(Struct) ->
    Struct0 = normalize(Struct),
    Struct00 = present_simp:cut(Struct0, 400),
    Struct000 = present_simp:eq(present_simp:fuse(Struct00)),
    Struct1=present_simp:cut(Struct000, 100),
    simpler(Struct1).

    
normalize(Struct) ->
    Struct2 = present_simp:reify(Struct),
    Struct3 = present_simp:purify(Struct2),
    present_simp:cleanup(Struct3).


simpler(Struct) ->
     Struct1 = present_simp:fuse(Struct),
     present_simp:equal(Struct1).


% Show struct

show_struct(H, S) ->
    io:put_chars(H, show:struct_to_string(S)).
  


%%% sketch prints a simple representation of the top-level structure
%%% of a term.


sketch(Term, AS) ->
    Term1 = sketch:trunc_term(Term, 5, AS),
    Tree = show:show_term(Term1),
    Hook = fun({trunc},_,_,_) ->
		   "..."
	   end,
    [erl_pp:exprs([Tree],0,Hook), "\n"].

%%% Generate_internal. When analysis is used in compiler


generate_internal(AS) ->
    Interesting = fun (Name) ->
			  tag:interesting(Name)
		  end,
    ContextTable = as:make_context_table(AS, Interesting),
    ContextList  = gb_trees:to_list(ContextTable),
    F = fun({Fnid, {ContextList2, NameList}}) ->
		gi_function(Fnid, ContextList2, NameList, AS)
	end,
    lists:append(lists:map(F,  ContextList)).

gi_function(Fnid, ContextList, NameList, AS) ->
    NameList1 = lists:sort(NameList),
    RootList = make_root_list(ContextList, NameList1, AS),
    Struct = present_struct:collect_roots(NameList1, RootList, AS),
    case program:deid_type(Fnid) of 
	fnid ->
	    MFA = {program:fnid_module(Fnid),
		   program:fnid_name(Fnid),
		   program:fnid_arity(Fnid)},
	    Struct2 = present_struct:rename(simplify(Struct)),
	    {AList, WList} = draw:draw_struct(Struct2),
	    [{function, MFA, AList, WList}];
	tyid ->
	    []
    end.


