%%% File    : sa_file.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : Interface between analyzer and file system
%%% Created :  7 Feb 2000 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(sa_file).
-author('svenolof@harpo.it.uu.se').

-export([find_module/3, normalize/1, check_path/1]).
-export([open_output_file/2,  open_output_file/3]).

% find_module looks for file with name M and extension Ext. Either
% fails with not_found, or returns {found, FileName}.

find_module(Path, Name, Ext) ->
    NameS = atom_to_list(Name),
    ExtS = atom_to_list(Ext),
    find_filename(Path, NameS, "."++ExtS).

find_filename([Dir | Path], Name, Suffix) ->
    Absname = filename:absname([Name,Suffix],Dir),
    case catch file:read_file_info(Absname) of
	{ok,_} -> 
	    {ok, Absname};
	_ ->
	    find_filename(Path, Name, Suffix)
    end;
find_filename([], _, _) ->
    not_found.

% open_output_file(Path, Name, Ext) creates a file for output in the
% first directory mentioned in path.

open_output_file(Name, Ext) ->
    [Dir | _] = path(),
    open_output_file(Dir, Name, Ext).

open_output_file(Dir, Name, Ext) ->
    NameS = atom_to_list(Name),
    ExtS = atom_to_list(Ext),
    Abs = filename:absname([NameS, "."++ExtS], Dir),
    {ok, H} = file:open(Abs, [write]),
    H.


%% Normalize the given path.
%% The check_path function catches erroneous path,
%% thus it is ignored here.

normalize([P|Path]) when atom(P) ->
    normalize([atom_to_list(P)|Path]);
normalize([P|Path]) when list(P) ->
    case int_list(P) of
	true -> [filename:join([P])|normalize(Path)];
	_    -> [P|normalize(Path)]
    end;
normalize([P|Path]) ->
    [P|normalize(Path)];
normalize([]) ->
    [];
normalize(Other) ->
    Other.

check_path([]) -> 
    true;
check_path([Dir |Tail]) ->
    case catch file:read_file_info(Dir) of
	{ok,_} -> 
	    check_path(Tail);
	_ -> 
	    {error, bad_directory}
    end;
check_path(_) ->
    {error, bad_path}.

int_list([H|T]) when integer(H) -> int_list(T);
int_list([_|T])                 -> false;
int_list([])                    -> true.


%%%%

%% Path --> moved to module program

path() ->
    case get(analysis_path) of
	undefined ->
	    ['.'];
	Path ->
	    Path
    end.



