-module(sketch).

% -compile(export_all).

-export([trunc_term/3]).



trunc_term(_, 0, _) ->
    term:trunc();
trunc_term(Term, N, AS) ->
    case term:term_tag(Term) of 
	functor ->
	    trunc_functor_term(Term, N, AS);
	'fun' ->
	    Term;
	ip ->
	    trunc_ip_ref_term(Term, N, AS);
	join ->
    	    trunc_join_term(Term, N, AS);
	universal ->
	    Term;
	parameter ->
	    Term;
	pid ->
	    term:make_pid_with_mbox({trunc})
    end.

trunc_functor_term(Term, N, AS) ->
    Functor = term:term_functor(Term),
    List = term:term_functor_arg_list(Term),
    trunc_functor(Functor, List, N, AS).


trunc_functor(atomc, [Term], _, _) ->
    case term:term_tag(Term) of
	universal ->
	    term:apply_functor(atomc,[Term]);
	functor ->
	    term:apply_functor(atomc,[Term])
    end;
trunc_functor(intc, [Term], _, _) ->
    case term:term_tag(Term) of
	universal ->
	    term:apply_functor(intc,[Term]);
	functor ->
	    term:apply_functor(intc,[Term])
    end;
trunc_functor(floatc, [Term], _, _) ->
    case term:term_tag(Term) of
	universal ->
	    term:apply_functor(float,[Term])
    end;
trunc_functor(tuple, [Term], N, AS) ->
     case is_clean_list(Term) of
	true ->
	     term:apply_functor(tuple, [trunc_clean_list(Term, N-1, AS)]);
	false ->
	     term:apply_functor(tuple, [trunc_term(Term, N-1, AS)])
    end;
trunc_functor(FunctorName, List, N, AS) ->
    List1 = trunc_list(List, N-1, AS),
    term:apply_functor(FunctorName, List1).
    

trunc_list([Term|TermList], N, AS) ->
    [trunc_term(Term, N, AS) |
     trunc_list(TermList, N, AS)];
trunc_list([], _, _) ->
    [].

is_clean_list(Term) ->
    case term:term_tag(Term) of
	functor ->
	    case term:term_functor(Term) of
		cons ->
		    [_, Tl] = term:term_functor_arg_list(Term),
		    is_clean_list(Tl);
		nil ->
		    true;
		_ ->
		    false
	    end;
	_ ->
	    false
    end.

trunc_clean_list(Term, N, AS) ->
    case term:term_tag(Term) of
	functor ->
	    case term:term_functor(Term) of
		cons ->
		    [Hd, Tl] = term:term_functor_arg_list(Term),
		    term:make_cons(trunc_term(Hd, N, AS),
				   trunc_clean_list(Tl, N, AS));
		nil ->
		    Term
	    end
    end.



trunc_ip_ref_term(Term, N, AS) ->
    Term1 = as:lookup(Term, AS),
    trunc_term(Term1, N, AS).

trunc_join_term(_, N, _) when N < 3 ->
    term:trunc();
trunc_join_term(Term, N, AS) ->
    case term:join_arg_list(Term) of
	[] ->
	    Term;
	[T] ->
	    trunc_term(T, N, AS);
	TermList ->
	    L = lists:map(fun(Term1) ->
				  trunc_term(Term1, N-1, AS)
			  end,
			  TermList),
	    term:join(L)		   
    end.
