%%% File    : spec.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : Handle specification files
%%% Created :  3 Nov 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

-module(spec).
-author('svenolof@harpo.csd.uu.se').

-export([spec_type/1, spec_functor/1, spec_functor_arg_list/1, make_functor/2,
	 typedef_name/1, typedef_arity/1, typedef_arg_list/1, typedef_body/1,
	 make_typedef/3,
%	 abstypedef_name/1, abstypedef_arg_list/1, abstypedef_body/1, 
	 make_abstypedef/3,  make_public_typedef/3,
	 funspec_name/1, funspec_arity/1, funspec_arg_list/1, funspec_guard/1,
	 funspec_result/1, funspec_line/1, make_funspec/5,
	 mfunspec_name/1, mfunspec_arity/1, mfunspec_clause_list/1,
	 mfunspec_line/1, make_mfunspec/4,
	 clause_arg_list/1, clause_result/1, clause_line/1, make_clause/3,
	 spec_form_type/1,
	 fun_arg_list/1, fun_result/1, fun_pos/1, make_fun/3,
	 pid_arg_list/1, make_pid_spec/0, make_pid_spec/1,
	 union_arg_list/1, make_union_spec/1,
	 var_name/1, make_var/1,
	 make_alias/2, alias_var_name/1, alias_spec/1,
	 make_universal/0,
	 apply_name/1, apply_arg_list/1, make_apply/3,
	 external_module/1, external_name/1, external_arg_list/1, make_external/4,
	 special_arity/1, special_name/1, make_special/2,
	 make_dict/2,
	 make_module_attribute/2, make_function_attribute/3, 
	 function_attribute_name_arity/1, attribute_tag/1, attribute_value/1,
	 pos/1]).

%%% The spec data type

spec_type(Spec) ->
%    io:format("~n Type: ~w",[Spec]),
    Type = element(1, Spec),
    case lists:member(Type, [var, alias, funspec, mfunspec, clause, functor, 'fun', 
			     pid, 
			     union, universal, empty, apply, external, special, dict,
			     module_attribute, function_attribute]) of
	true ->
	    Type
    end.

spec_functor({functor, F, _}) ->
    F.

spec_functor_arg_list({functor, _, ArgList}) ->
    ArgList.

make_functor(F, ArgList) ->
    {functor, F, ArgList}.

typedef_name({typedef, _, Name, ArgList, Body}) ->
    Name.

typedef_arg_list({typedef, _, Name, ArgList, Body}) ->
    ArgList.

typedef_arity(Typedef) ->
    length(typedef_arg_list(Typedef)).

typedef_body({typedef, _, Name, ArgList, Body}) ->
    Body.

make_typedef(Name, ArgList, Body) ->
    {typedef, normal, Name, ArgList, Body}.

make_abstypedef(Name, ArgList, Body) ->
    {typedef, abs, Name, ArgList, Body}.

make_public_typedef(Name, ArgList, Body) ->
    {typedef, public, Name, ArgList, Body}.



funspec_name({funspec, Name, ArgList, Guard, Body, Line}) ->
    Name.

funspec_arity({funspec, Name, ArgList, Guard, Body, Line}) ->
    length(ArgList).

funspec_arg_list({funspec, Name, ArgList, Guard, Body, Line}) ->
    ArgList.

funspec_guard({funspec, Name, ArgList, Guard, Body, Line}) ->
    Guard.

funspec_result({funspec, Name, ArgList, Guard, Body, Line}) ->
    Body.

funspec_line({funspec, Name, ArgList, Guard, Body, Line}) ->
    Line.


make_funspec(Name, ArgList, Guard, Body, Line) ->
    {funspec, Name, ArgList, Guard, Body, Line}.

mfunspec_name({mfunspec, Name, Arity, ClauseList, Line}) ->
    Name.

mfunspec_arity({mfunspec, Name, Arity, ClauseList, Line}) ->
    Arity.

mfunspec_clause_list({mfunspec, Name, Arity, ClauseList, Line}) ->
    ClauseList.

mfunspec_line({mfunspec, Name, Arity, ClauseList, Line}) ->
    Line.

make_mfunspec(Name, Arity, ClauseList, Line) ->
    {mfunspec, Name, Arity, ClauseList, Line}.

clause_arg_list({clause, ArgList, Result, Line}) ->
    ArgList.    

clause_result({clause, ArgList, Result, Line}) ->
    Result.

clause_line({clause, ArgList, Result, Line}) ->
    Line.


make_clause(ArgList, Result, Line) ->
    {clause, ArgList, Result, Line}.

spec_form_type({typedef, normal, Name, ArgList, Body}) ->
    typedef;
spec_form_type({typedef, abs, Name, ArgList, Body}) ->
    abstypedef;
spec_form_type({typedef, public, Name, ArgList, Body}) ->
    public_typedef;
spec_form_type({funspec, Name, ArgList, Guard, Result, Line}) ->
    funspec;
spec_form_type({mfunspec, Name, Arity, ClauseList, Line}) ->
    mfunspec;
spec_form_type({special, _, _}) ->
    special;
spec_form_type({dict, _, _}) ->
    dict;
spec_form_type({file, _}) ->
    file;
spec_form_type({module, _}) ->
    module;
spec_form_type({eof, _}) ->
    eof;
spec_form_type({module_attribute, _, _}) ->
    module_attribute;
spec_form_type({function_attribute, _, _, _}) ->
    function_attribute.
    
%%%%

make_fun(ArgList, Result, Pos) ->
    {'fun', ArgList, Result, Pos}.

fun_arg_list({'fun', ArgList, _, _}) ->
    ArgList.

fun_result({'fun', _, Result, _}) ->
    Result.

fun_pos({'fun', _, _, Pos}) ->
    Pos.

% Pid

make_pid_spec() ->
    {pid}.

make_pid_spec(Type) ->
    {pid, Type}.

pid_arg_list({pid}) ->
    [];
pid_arg_list({pid, Type}) ->
    [Type].

% Union

make_union_spec(L) ->
    {union, L}.

union_arg_list({union, L}) ->
    L.

var_name({var, Var}) ->
    Var.

make_var(Name) ->
    {var, Name}.

make_alias(Name, Type) ->
    {alias, Name, Type}.

alias_var_name({alias, Name, Type}) ->
    Name.

alias_spec({alias, Name, Type}) ->
    Type.

make_universal() ->
    {universal}.

% Apply

make_apply(Name, ArgList, Pos) ->
    {apply, Name, ArgList, Pos}.

apply_name(Apply) ->
    element(2, Apply).

apply_arg_list({apply, _, ArgList, _} ) ->
    ArgList.

clean_apply({apply, Name, ArgList, _}) ->
    {apply, Name, ArgList, 0}.

make_external(Module, Name, ArgList, Pos) ->
    {external, Module, Name, ArgList, Pos}.

external_module(Apply) ->
    element(2, Apply).

external_name(Apply) ->
    element(3, Apply).

external_arg_list({external, _, _, ArgList, _} ) ->
    ArgList.

make_special(Name, Arity) ->
    {special, Name, Arity}.

special_name({special, Name, Arity}) ->
    Name.

special_arity({special, Name, Arity}) ->
    Arity.

make_dict(Key, Type) ->
    {dict, Key, Type}.

dict_key({dict, Key, Type}) ->
    Key.

dict_type({dict, Key, Type}) ->
    Type.

% 

make_module_attribute(Tag, Type) ->
    {module_attribute, Tag, Type}.

make_function_attribute(Tag, LFnid, Type) ->
    {function_attribute, Tag, LFnid, Type}.

function_attribute_name_arity({function_attribute, Tag, {lfnid, Name, Arity}, Type}) ->
    {Name, Arity}.

attribute_tag({module_attribute, Tag, Type}) ->
    Tag;
attribute_tag({function_attribute, Tag, LFnid, Type}) ->
    Tag.

attribute_value({module_attribute, Tag, Type}) ->
    Type;
attribute_value({function_attribute, Tag, LFnid, Type}) ->
    Type.



% Exception

make_exception_spec(T) ->
    {exception, T}.

% Position


pos({apply, _Name, _ArgList, Pos}) ->
    Pos;
pos({external, _Module, _Name, _ArgList, Pos}) ->
    Pos;
pos({'fun', _, _, Pos}) ->
    Pos.
