%% File    : spec_parse.yrl
%% Author  : Sven-Olof Nystrom (adapted from erl_parse)
%% Purpose : Type syntax

Nonterminals
form
attribute attr_val
type_decl type_header
utype_list utype_tuple utypes utype ptypes ptype funtype  qclauses
function 
expr 
argument_list
exprs
atomic
function_clause function_clauses
special_entry dict_entry.

Terminals
atom float integer string var

'(' ')' ',' '->' 
'{' '}' '[' ']' '|' ':' ';'
'end' 'fun' 

'-' '+' '/'
'='
dot.

Rootsymbol form.


form -> attribute dot : '$1'.
form -> type_decl dot : '$1'.
form -> function dot : '$1'.
form -> special_entry dot : '$1'.
form -> dict_entry dot : '$1'.
    

attribute -> '-' atom '(' attr_val ')' : build_attribute('$2', '$4').

attr_val -> exprs : '$1'.

type_decl -> '-' atom type_header '=' utype :
	if
	    element(3, '$2') == type ->
		{type,element(2, '$2'),def,'$3','$5'};
	    element(3, '$2') == abstype ->
		{abstype,element(2, '$2'),def,'$3','$5'};
	    element(3, '$2') == public_type ->
		{public_type,element(2, '$2'),def,'$3','$5'};
	    true ->
		return_error(element(2,'$2'), "illegal type declaration")
	end.

type_header -> atom utype_list : {element(3, '$1'),'$2'}.


utype_list -> '(' ')' : [].
utype_list -> '(' utypes ')' : '$2'.

utype_tuple -> '{' '}' : [].
utype_tuple -> '{' utypes '}' : '$2'.

utypes -> utype : ['$1'].
utypes -> utype ',' utypes : ['$1'|'$3'].

utype -> ptypes : {utype,'$1'}.


ptypes -> ptype : ['$1'].
ptypes -> ptypes '+' ptype : '$1' ++ ['$3'].
ptypes -> ptypes '|' ptype : io:format("~nObsolete specification syntax!",[]), '$1' ++ ['$3'].

ptype -> var : element(3, '$1').
ptype -> var '=' ptype : {alias,  element(3, '$1'), '$3'}.
ptype -> '(' utype ')' : '$2'.
ptype -> '(' ')' : {utype,[]}.
ptype -> atom utype_list : {type,element(3, '$1'),'$2'}.
ptype -> atom ':' atom utype_list : 
	{external, element(3, '$1'),element(3, '$3'),'$4'}.
ptype -> atom : {atom, element(2, '$1'), element(3, '$1')}.
ptype -> utype_tuple : {tuple,'$1'}.
ptype -> '[' ptype '|' ptype ']' : {cons, element(2, '$1'), '$2', '$4'}.
ptype -> '[' ']' : {nil, element(2, '$1')}.
ptype -> integer : {int, '$1'}.
ptype -> funtype : '$1'.

qclauses ->  utype '->' utype : [{'$1', '$3'}].
qclauses -> utype '->' utype ';' qclauses : [{'$1', '$3'} | '$5'].
    
    
    
    
    
funtype ->
    'fun' utype_list '->' utype 'end' : {'fun', '$2', '$4', element(2, '$1')}.

function -> function_clauses : build_function('$1').
%function -> function_clauses '&' function_clause : build_function('$1', '$3').

function_clauses -> function_clause : ['$1'].
function_clauses -> function_clause ';' function_clauses : ['$1'|'$3'].

function_clause -> atom utype_list '->' utype :
	{clause, element(2, '$1'), element(3, '$1'),'$2', [], '$4'}.

expr -> var : '$1'.
expr -> atomic : '$1'.
expr -> atom '/' integer : {lfnid, element(3, '$1'), element(3, '$3')}.


argument_list -> '(' ')' : {[],element(2, '$1')}.
argument_list -> '(' exprs ')' : {'$2',element(2, '$1')}.

exprs -> expr : ['$1'].
exprs -> expr ',' exprs : ['$1' | '$3'].

atomic -> integer : '$1'.
atomic -> float : '$1'.
atomic -> atom : '$1'.
atomic -> string : '$1'.

special_entry -> '+' atom '=' utype : 
		     {special_entry, term('$2'), '$4'}.
special_entry -> '+' atom '(' expr ')' '=' utype : 
		     {special_entry, term('$2'), term('$4'), '$7'}.

dict_entry -> '+' atom '=' '(' qclauses ')' :
		  {dict_entry, term('$2'), '$5'}.
dict_entry -> '+' atom '(' expr ')' '=' '(' qclauses ')' :
		  {dict_entry, term('$2'), term('$4'), '$8'}.
    


Erlang code.


-export([parse_form/1]).

parse_form(Tokens) ->
     parse(Tokens).


build_attribute({atom,La,module}, Val) ->
    case Val of
	[{atom,Lm,Module}] -> {attribute,La,module,Module};
	Other -> return_error(La, "bad module declaration")
    end;
build_attribute({atom,La,file}, Val) ->
    case Val of
	[{string,Ln,Name},{integer,Ll,Line}] ->
	    {attribute,La,file,{Name,Line}};
	Other -> return_error(La, "bad file declaration")
    end
%;
% build_attribute({atom,La,Attr}, Val) ->
%     case Val of
% 	[Expr] ->
% 	    io:format("~n~n Attribute : ~w~n~n",[Expr]),
% 	    {attribute,La,Attr,Expr};
% 	[E1, E2] ->
% 	    {attribute2, La, Attr, term(E1), E2};
% 	Other -> return_error(La, "bad attribute")
%    end
.

term(Expr) ->
    case catch normalise(Expr) of
	{'EXIT',R} -> return_error(element(2, Expr), "bad attribute");
	Term -> Term
    end.


%% build_function([Clause]) -> {function,Line,Name,Arity,[Clause], Otherwise}

build_function(Cs) ->
    Name = element(3, hd(Cs)),
    Arity = length(element(4, hd(Cs))),
    {function,element(2, hd(Cs)),Name,Arity,check_clauses(Cs, Name, Arity)}.

check_clauses(Cs, Name, Arity) ->
     lists:map(fun ({clause,L,N,As,G,B}) when N == Name, length(As) == Arity ->
		 {clause,L,As,G,B};
	     ({clause,L,N,As,G,B}) ->
		 return_error(L, "head mismatch") end, Cs).


normalise({atom,_,A}) -> A;
normalise({lfnid, F, N}) -> {lfnid, F, N}.




