%%% File    : strata.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : 
%%% Created : 29 Oct 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

-module(strata).
-author('svenolof@harpo.csd.uu.se').

-export([init/0, setup_module/1, setup_module_list/1]).


init() ->
    put(strata_table, gb_trees:empty()),
    true.

setup_module(M) ->
    T00 = get(strata_table),
    G = flow:module_flow_graph(M),
    T = compute_strata(G, T00),
    put(strata_table, T),
    WT = compute_weight_table(G, T),
    put(weight_table, WT),
    true.

setup_module_list(MList) ->
    T00 = get(strata_table),
    G = flow:multi_module_flow_graph(MList),
    T = compute_strata(G, T00),
    put(strata_table, T),
    WT = compute_weight_table(G, T),
    put(weight_table, WT),
    true.
    

compute_strata(G, StrataTable) ->
    T00 = StrataTable,
    CycleList = graph:cycles(G),
    lists:foldl(fun(Cycle, T0) ->
                        Fnid = hd(Cycle),
                        lists:foldl(fun(Fnid1, T1) ->
                                            gb_trees:insert(Fnid1, Fnid, T1)
                                    end,
                                    T0,
                                    Cycle)
                end,
                T00,
                CycleList).

stratum(Node, Strata) ->
   {value, S} = gb_trees:lookup(Node, Strata),
    S.

make_dag(FlowGraph, Strata) ->
    G0 = graph:empty_graph(),
    F = fun(Father, Son, Weight, DAG) ->
		Father1 = stratum(Father, Strata),
		Son1 = stratum(Son, Strata),
		if Father1==Son1 ->
			graph:new_node(Father1, DAG);
		   true  ->
			graph:insert_weighted_son(Father1, Son1, Weight, DAG)
		end
	end,
    graph:foldl(F, G0, FlowGraph).

node_weight(Node, Table, DAG) ->
    case gb_trees:lookup(Node, Table) of
	{value, Weight} ->
	    {Weight, Table};
	none ->
	    SonWeightList = graph:son_weight_list(Node, DAG),
	    {Weight, Table1} = node_weight_list(SonWeightList, 1, Table, DAG),
	    {Weight, gb_trees:insert(Node, Weight, Table1)}
    end.

node_weight_list([{_, 0} | List], ACC, Table, DAG) ->
    node_weight_list(List, ACC, Table, DAG);
node_weight_list([{Node, EdgeWeight} | List], ACC, Table, DAG) ->
    {Weight, Table1} = node_weight(Node, Table, DAG),
    node_weight_list(List, ACC+EdgeWeight*Weight, Table1, DAG);
node_weight_list([], ACC, Table, _) ->
    {ACC, Table}.
    
compute_weight_table(FlowGraph, Strata) ->
    DAG = make_dag(FlowGraph, Strata),
%     io:format("~n~n== DAG ==",[]),
%     show_dag(DAG),
    NodeList = graph:node_list(DAG),
    F = fun(Node, Table) ->
		{_, Table1} = node_weight(Node, Table, DAG),
		Table1	
	end,
    lists:foldl(F, gb_trees:empty(), NodeList).

%% show_dag(Graph) ->
%%     F= fun({Node, Sons}) ->
%% 	       SonList = gb_trees:to_list(Sons),
%% 	       io:format("~nNode ~w => ~w ", [Node, SonList])
%%        end,
%%     lists:foreach(F, gb_trees:to_list(Graph)).

