%%% File    : tag.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : 
%%% Created : 15 Aug 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

-module(tag).
-author('svenolof@harpo.csd.uu.se').

-export([self/1, except/1, result/1, 
	 arg/2,
	 fun_arg/3,
	 fun_result/2,
	 match_result/1,
	 vector/3,
	 var/3,
	 try_except/2,
	 try_body_result/2,
	 call_result/2, 
	 receive_result/2,
	 apply_result/2,
	 any_string/0,
	 spawn_mbox/2, spawn_result/2, spawn_except/2,
	 pid_dict/2, node_registry/1,
	 nth_result/2,
	 set_nth_result/2,
	 interesting/1]).
-export([record_bif/4, bif_call/1]).

-define(proper, element(4,Context)).
% See definition of function context:context_proper.

self(Context) ->
    term:make_ip(self, ?proper).

except(Context) ->
    term:make_ip(except, ?proper).

arg(N, Context) ->
    Fnid = context:current_fnid(Context),
    term:make_ip({arg, N, Fnid}, ?proper).


result(Context) ->
    Fnid = context:current_fnid(Context),
    term:make_ip({result, Fnid}, ?proper).



apply_result(Pos, Context) ->
    term:make_ip({apply_result, Pos}, ?proper).

call_result(Pos, Context) ->
    term:make_ip({call_result, Pos}, ?proper).



fun_result(Context, Pos) ->
    term:make_ip({fun_result, Pos}, ?proper).


fun_arg(Context, N, Pos) ->
    term:make_ip({fun_arg, Pos, N}, ?proper).


match_result(Context) ->
    term:make_ip(match_result, ?proper).

vector(Context, N, Pos) ->
    vector_1(?proper, 1, N, Pos).

vector_1(PC, I, N, Pos) when I =< N ->
    [vector_index(PC, I, Pos) | 
     vector_1(PC, I+1, N, Pos)];
vector_1(_Context, I, N, _Pos) when I>N ->
    [].

vector_index(PC, N, Pos) ->
    term:make_ip({vector_index, N, Pos}, PC).


var(Name, Pos, Context) ->
    term:make_ip({var, Name, Pos}, ?proper).


spawn_result(Pos, Context) ->
    term:make_ip({spawn_result, Pos}, ?proper).


spawn_mbox(Pos, Context) ->
    term:make_ip({spawn_mbox, Pos}, ?proper).

pid_dict(Pos, Context) ->
    term:make_ip({pid_dict, Pos}, ?proper).

spawn_except(Pos, Context) ->
    term:make_ip({spawn_except, Pos}, ?proper).

try_except(Pos, Context) ->
    term:make_ip({try_except, Pos}, ?proper).

try_body_result(Pos, Context) ->
    term:make_ip({try_body_result, Pos}, ?proper).


receive_result(Pos, Context) ->
    term:make_ip({receive_result, Pos}, ?proper).


node_registry(Context) ->
    term:make_ip(registry, ?proper).


any_string() ->
    term:make_ip(any_string, context:empty()).

nth_result(Context, Pos) ->
    term:make_ip({nth_result, Pos}, ?proper).

set_nth_result(Context, Pos) ->
    term:make_ip({set_nth_result, Pos}, ?proper).

%%% The presentation interface needs to know whether a tag is
%%% interesting!

interesting({arg, _, Fnid}) ->
    {ok, Fnid};
interesting({result, Fnid}) ->
    {ok, Fnid};
interesting(self) ->
    false;
interesting(except) ->
    false;
interesting(any_string) ->
    false;
interesting(_) ->
    false.

%%% To measure precision of analysis, we record for each call to a
%%% bif: the fnid of the caller, the fnid of the bif, the value of
%%% each argument.

record_bif(Caller, Bif, Pos, K) ->
    term:make_ip({record_bif, Caller, Bif, Pos, K}, context:call(Caller, 0, context:empty())).

bif_call({record_bif, Caller, _, _, _}) ->
    {ok, Caller};
bif_call(_) ->
    false.
