%%% File    : top_level.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : Top-level of analysis
%%% Created :  3 Jul 2000 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(top_level).
-author('svenolof@harpo.it.uu.se').

%%-compile(export_all).
-export([top_level/3]).


top_level(Fnid, Context, AS) ->
    {spec, Spec, Env} = program:spec(Fnid),
    case spec:spec_type(Spec) of 
	funspec ->
	    top_level_one(Fnid, Context, AS);
	mfunspec ->
	    top_level_many(Fnid, Context, AS)
    end.
    
    
top_level_one(Fnid, Context, AS) ->
    {spec, Spec, Env} = program:spec(Fnid),
    SpecArgList = spec:funspec_arg_list(Spec),
    AS0 = AS,
    SpecResult = spec:funspec_result(Spec) ,
    Arity =  length(SpecArgList),
    Env1 = bind_parameters(Spec, Env, Context),
    {ArgList, AS1} = spec_generate:generate_list(SpecArgList,  Env1, Context, AS0),
    {CallContext, AS2} = work:call_analysis(Fnid, 0, Context, AS1),
    AS4 = param:insert_arg_list(CallContext, ArgList, AS2),
    TermResult = tag:result(CallContext),
    case spec_match:match(SpecResult, TermResult, Env1, Context, AS4) of
	{ok, AS8} ->
	    true;
	{fail, AS7} ->
	    walker:set_current_fnid(Fnid),
	    AS8 = report:report_bad_result(Fnid, TermResult, AS7)
    end,
    Cont = fun(_, BS) ->
		   report:report_bad_result(Fnid, TermResult, BS)
	   end,
    as:lookup_cont(tag:match_result(Context), Cont, Context, AS8).

top_level_many(Fnid, Context, AS) ->
    {spec, Spec, Env} = program:spec(Fnid),
    ClauseList = spec:mfunspec_clause_list(Spec),
    F = fun(Clause, S) -> top_level_clause(Clause, Fnid, Env, Context, S) end,
    lists:foldl(F, AS, ClauseList).

top_level_clause(Clause, Fnid, Env, Context, AS) ->
    io:format("~nClause ~p",[Clause]),
    Env1 = bind_parameters(Clause, Env, Context),
    SpecArgList = spec:clause_arg_list(Clause),
    SpecResult = spec:clause_result(Clause) ,
    {ArgList, AS1} = spec_generate:generate_list(SpecArgList,  Env1, Context, AS),
    {CallContext, AS2} = work:call_analysis(Fnid, 0, Context, AS1),
    AS4 = param:insert_arg_list(CallContext, ArgList, AS2),
    TermResult = tag:result(CallContext),
      case spec_match:match(SpecResult, TermResult, Env1, Context, AS4) of
	{ok, AS7} ->
	    true;
	{fail, AS6} ->
	      walker:set_current_fnid(Fnid),
	      AS7 = report:report_bad_result(Fnid, TermResult, AS6)
    end,
    Cont = fun(_, BS) ->
		   report:report_bad_result(Fnid, TermResult, BS)
	   end,
    as:lookup_cont(tag:match_result(Context), Cont, Context, AS7).

    
bind_parameters(Spec, Env, Context) ->
    case spec:spec_type(Spec) of
	funspec ->
	    ArgList = spec:funspec_arg_list(Spec),
	    Result = spec:funspec_result(Spec),
	    bind_parameter_list([Result | ArgList], Env, Context);
	functor ->
	    bind_parameter_list(spec:spec_functor_arg_list(Spec), Env, Context);
	'fun' ->
	    ArgList = spec:fun_arg_list(Spec),
	    Result = spec:fun_result(Spec),
	    bind_parameter_list([Result | ArgList], Env, Context);
	pid ->
	    ArgList = spec:pid_arg_list(Spec),
	    bind_parameter_list(ArgList, Env, Context);
	union ->
	    List = spec:union_arg_list(Spec),
	    bind_parameter_list(List, Env, Context);
	universal ->
	    Env;
	var ->
	    Name = spec:var_name(Spec),
	    case env:defined(Name, Env) of
		true ->
		    Env;
		false ->
		    env:bind(Name, term:parameter(Name), Env)
	    end;
	alias ->
	    Name = spec:alias_var_name(Spec),
	    case env:defined(Name, Env) of
		true ->
		    Env1 = Env;
		false ->
		    Env1 = env:bind(Name, term:parameter(Name), Env)
	    end,
	    Spec1 = spec:alias_spec(Spec),
	    bind_parameters(Spec1, Env1, Context);
	apply ->
	    ArgList = spec:apply_arg_list(Spec),
	    bind_parameter_list(ArgList, Env, Context);
	external ->
	    ArgList = spec:external_arg_list(Spec),
	    bind_parameter_list(ArgList, Env, Context);
	clause ->
	    ArgList = spec:clause_arg_list(Spec),
	    Result = spec:clause_result(Spec),
	    bind_parameter_list([Result|ArgList], Env, Context)
    end.

bind_parameter_list([Spec | SpecList], Env, Context) ->
    Env1 = bind_parameters(Spec, Env, Context),
    bind_parameter_list(SpecList, Env1, Context);
bind_parameter_list([], Env, _) ->
    Env.



