%%% File    : wait.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : 
%%% Created : 30 Oct 2000 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(wait).
-author('svenolof@harpo.it.uu.se').

-export([wait/5, wait_vector_r/4]).

wait([Term | TermList], [Shape | ShapeList], F, Context, AS) ->
    G = fun(Term1, BS) ->
		H = fun(TermList1, CS) ->
			    F([Term1 | TermList1], CS)
		    end,
		wait(TermList, ShapeList, H, Context, BS)
	end,
    wait_shape(Term, Shape, G, Context, AS);
wait([], [], F, Context, AS) ->
    F([], AS).


wait_shape(Term, atom, F, Context, AS) ->
    wait_atom(Term, F, Context, AS);
wait_shape(Term, pid, F, Context, AS) ->
    wait_pid(Term, F, Context, AS);
wait_shape(Term, {list, N}, F, Context, AS) ->
    wait_list(Term, N, F, Context, AS);
wait_shape(Term, {functor, N}, F, Context, AS) ->
    wait_list(Term, N, F, Context, AS);
wait_shape(Term, universal, F, Context, AS) ->
    wait_universal(Term, F, Context, AS);
wait_shape(Term, vector, F, Context, AS) ->
    wait_vector(Term, F, Context, AS).

wait_atom(Term, F, Context, AS) ->
    case term:term_tag(Term) of
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    wait_atom(Term1, F, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    wait_atom_list(TermList, F, Context, AS);
	functor ->
	    case term:is_atom(Term) of
		true ->
		    F(Term, AS);
		false ->
		    AS
	    end;
	universal ->
	    F(Term, AS);
	_ ->
	    AS
    end.

wait_atom_list([Term | TermList], F, Context, AS) ->
    AS1 = wait_atom(Term, F, Context, AS),
    wait_atom_list(TermList, F, Context, AS1);
wait_atom_list([], _, _, AS) ->
    AS.	

wait_pid(Term, F, Context, AS) ->
    case term:term_tag(Term) of
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    wait_pid(Term1, F, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    wait_pid_list(TermList, F, Context, AS);
	pid ->
	    F(Term, AS);
	universal ->
	    F(Term, AS);
	_ ->
	    AS
    end.

wait_pid_list([Term | TermList], F, Context, AS) ->
    AS1 = wait_pid(Term, F, Context, AS),
    wait_pid_list(TermList, F, Context, AS1);
wait_pid_list([], _, _, AS) ->
    AS.	

wait_list(L, N, F, Context, AS) ->
    G = fun (Term, ASa) ->
		case term:is_cons(Term) of
		    true when N > 0 ->
			Hd = term:cons_car(Term),
			Tl = term:cons_cdr(Term),
			H = fun (TermList, ASq) ->
				    F([Hd | TermList], ASq)
			    end,
			wait_list(Tl, N-1, H, Context, ASa);
		    _ ->
			case term:is_nil(Term) of
			    true when N == 0 ->
				F([], ASa);
			    _ ->
				case term:term_tag(Term) == universal of
				    true ->
					H = fun (TermList, ASq) ->
						    F([Term | TermList], ASq)
					    end,
					wait_list(Term, N-1, H, Context, ASa);
				    _ ->
					ASa
				end
			end
		end
	end,
    AS1 = term:do_all(L, G, Context, AS).

wait_universal(Term, F, Context, AS) ->
    case term:term_tag(Term) of
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    wait_universal(Term1, F, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    wait_universal_list(TermList, F, Context, AS);
	_ ->
	    F(Term, AS)
    end.

wait_universal_list([Term | TermList], F, Context, AS) ->
    AS1 = wait_universal(Term, F, Context, AS),
    wait_universal_list(TermList, F, Context, AS1);
wait_universal_list([], _, _, AS) ->
    AS.

wait_vector(Term, F, Context, AS) ->
    case term:term_tag(Term) of
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    wait_vector(Term1, F, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    wait_vector_list(TermList, F, Context, AS);
	functor ->
	    case term:is_vector(Term) of
		true ->
		    TermList = term:vector_arg_list(Term),
		    F(TermList, AS);
		false ->
		    F([Term], AS)
	    end;
	_ ->
	    F([Term], AS)
    end.


wait_vector_list([Term | TermList], F, Context, AS) ->
    AS1 = wait_vector(Term, F, Context, AS),
    wait_vector_list(TermList, F, Context, AS1);
wait_vector_list([], _, _, AS) ->
    AS.

% wait_*_r assumes that F returns a pair {Term, AS} and will also
% return a pair {Term, AS}.

wait_vector_r(Term, DestList, Context, AS) ->
    wait_vector_r(Term, DestList, Context, AS, 0).

wait_vector_r(Term, DestList, Context, AS,N) ->
    case term:term_tag(Term) of
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    AS2 = wait_vector_r(Term1, DestList, Context, AS1, N+1),
	    AS2;
	join ->
	    TermList = term:join_arg_list(Term),
	    AS1 = wait_vector_list_r(TermList, DestList, Context, AS, N+1),
	    AS1;
	functor ->
	    case term:is_vector(Term) of
		true ->
		    TermList = term:vector_arg_list(Term),
		    insert_list(DestList, TermList, AS);
		false ->
		    insert_list(DestList, [Term], AS)
	    end;
	_ ->
	    insert_list(DestList, [Term], AS)
    end.


wait_vector_list_r(TermList, DestList, Context, AS, N) ->
    wait_vector_list_r_1(TermList, DestList, Context, AS, N).

wait_vector_list_r_1([Term | TermList], DestList, Context, AS, N) ->
    AS1 = wait_vector_r(Term, DestList, Context, AS, N+1),
    AS2 = wait_vector_list_r_1(TermList, DestList, Context, AS1, N),
    AS2;
wait_vector_list_r_1([], _, _, AS, _) ->
    AS.	

insert_list(DestList, TermList, AS) 
  when length(TermList) == length(DestList) ->
    insert_list_1(DestList, TermList, AS).

insert_list_1([Dest | DestList], [Term | TermList], AS) ->
    AS1 = as:insert(Dest, Term, AS),
    insert_list_1(DestList, TermList, AS1);
insert_list_1([], [], AS) ->
    AS.


