%%% File    : work.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : 
%%% Created : 27 Apr 2000 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(work).
-author('svenolof@harpo.it.uu.se').

-export([iter/1, current_work/0,
	 call_top_level/4, 
	 call_analysis/4, 
	 call_matching/4,
	 call_generation/4,
	 call_specification/4,
	 work_index/1]).

%%% Iteration

iter(AS) ->
    case as:get_work(AS) of
	{ok, Work, AS2} ->
	    AS3 = work(Work, AS2),
	    iter(AS3);
	undefined  ->
	    AS
    end.


%%% Analysis points

work(Work, AS) ->
    set_current_work(Work),
    case Work of
	{top_level, AP} ->
	    Fnid = context:current_fnid(AP),
	    walker:set_current_fnid(Fnid),
	    top_level:top_level(Fnid, AP, AS);
	{analysis, AP} ->
	    Fnid = context:current_fnid(AP),
	    note(Fnid),
	    Context = AP,
	    walker:set_current_fnid(Fnid),
	    Definition = program:def(Fnid),
	    walker:definition(Definition, Context, AS);
	{specification,  AP} ->
	    Fnid = context:current_fnid(AP),
	    walker:set_current_fnid(Fnid),
	    {spec, Type, Env} = program:spec(Fnid),
	    spec_walker:specification(Fnid, Type, Env, AP, AS);
	{generation, AP} ->
	    Tyid = context:current_tyid(AP),
	    Form =  program:typedef(Tyid),
	    spec_walker:generate_typedef(Form, AP, AS);
	{matching, AP} ->
	    Tyid = context:current_tyid(AP),
	    Form = program:typedef(Tyid),
	    spec_walker:match_typedef(Form, AP, AS)
    end.

current_work() ->
    get(current_work).

set_current_work(Name) ->
    put(current_work, Name).

%%% Calls. Intended to replace a lot of other stuff

call_top_level(Fnid, Pos, Context, AS) ->
    AP = context:call(Fnid, Pos, Context),
    AS1 = as:add_visited({top_level, AP}, Context, AS),
    {AP, AS1}.

call_analysis(Fnid, Pos, Context, AS) ->
    AP = context:call(Fnid, Pos, Context),
    AS1 = as:add_visited({analysis, AP}, Context, AS),
    {AP, AS1}.

call_matching(Fnid, Pos, Context, AS) ->
    AP = context:call(Fnid, Pos, Context),
    AS1 = as:add_visited({matching, AP}, Context, AS),
    {AP, AS1}.

call_generation(Fnid, Pos, Context, AS) ->
    AP = context:call(Fnid, Pos, Context),
    AS1 = as:add_visited({generation, AP}, Context, AS),
    {AP, AS1}.

call_specification(Fnid, Pos, Context, AS) ->
    AP = context:call(Fnid, Pos, Context),
    AS1 = as:add_visited({specification, AP}, Context, AS),
    {AP, AS1}.

%%

work_index({W, Context}) ->
     {W, context:current_fnid(Context), context:context_proper(Context)}.

note(_) ->
	true.

