-module(erlang).

-type list(X) = [] + [X | list(X)].
-type char() = int().
-type number() = int() + float().
-type bool() = true + false.

-abstype binary() = binary().

-abstype ref() = ref().

%%% 2. Atoms

atom_to_list(atom()) ->
    list(char()).

list_to_atom(list(char())) ->
    atom().

%%% 3. Arithmetic

is_number(_) -> bool().

abs(int()) ->
    int();
abs(float()) ->
    float().

float(number()) ->
    float().

float_to_list(float()) ->
    list(char()).

list_to_float(list(char())) ->
	float().

integer_to_list(int()) ->
    list(char()).

list_to_integer(list(char())) ->
    int().

'div'(int(), int()) ->
    int().

'rem'(int(), int()) ->
    int().

round(number()) ->
    int().

trunc(number()) ->
    int().


%%% 5. Tuples


% element(int(), tuple(list(T))) ->
%     T.

element(N = int(), tuple(L)) ->
    nth(N, L).

setelement(N=int(), tuple(L = list(_)), T) ->
    tuple(set_nth(N, L, T)).




% element(N = int(), tuple(L= list(_))) ->
%     nth(N, L).

% setelement(N=int(), tuple(L = list(_)), T) ->
%     tuple(set_nth(N, L, T)).

list_to_tuple(L) ->
    tuple(L).

tuple_to_list(tuple(L)) ->
    L.

size(_) ->
    int().

%%% 6. Lists and conses

length(list(_)) ->
    int().

hd([X|_]) ->
    X;
hd([]) ->
    none().

tl([_|X]) ->
    X;
tl([]) ->
    none().

%%% 8. Functions and processes

is_builtin(atom(), atom(), int()) ->
    bool().

apply(Module, Name, ArgList) ->
    special_definition().

apply({Module, Name}, ArgList) ->
    apply(Module, Name, ArgList).

% {def, apply, [module, name, arglist],
%  {apply, module, name, arglist,
%   result, self, except}}.

% {def, exit, [message],
%  {insert, except, {tuple, [{make_atom, 'EXIT'}, message]}}}.

exit(_) ->
    special_definition().

exit(_, _) ->
    true. % Should also send exit message.

exit_current(_) ->
    special_definition().

fault(_) ->
    special_definition().

fault(_, _) ->
    special_definition().

group_leader() ->
    pid().

halt() ->
    none().

process_flag(_, _) ->
    any().

throw(_) ->
    special_definition().

self() ->
    special_definition().

spawn(_,_,_) ->
    special_definition().

spawn_opt(_,_,_,_) ->
    special_definition().

spawn_link(_,_,_) ->
    special_definition().

spawn(_) ->
    special_definition().

'!'(_,_) ->
    special_definition().

link(_) ->
    true.

unlink(_) ->
    true.

pid_to_list(pid()) ->
    list(char()).

% {def, self, [],
%  {insert, result, self}}.

% {def, spawn,  [module, name, arglist],
%  [{bind, node, {pid_node, self}},
%   {bind, new_result, {tag, spawn_result, [pos]}},
%   {bind, new_self,   {make_pid, pos, node}},
%   {bind, new_except, {tag, spawn_except, [pos]}},
%   {insert, result, new_self},
%   {apply, module, name, arglist,
%    new_result, new_self, new_except}]}.

% {def,'!', [dest, message],
%  {'or', 
%   [{bind, mbox, {pid_mbox, dest}},
%    {insert, mbox, message},
%    {insert, result, message}
%   ],
%   [{bind, node, {pid_node, self}},
%    {bind, registry, {check_cons, {node_registry, node}}},
%    {test, {eq, {car, registry}, {check_atom, dest}}},
%    {bind, process, {cdr, registry}},
%    {bind, mbox, {pid_mbox, process}},
%    {insert, mbox, message},
%    {insert, result, message}]}}.

% {def, group_leader, [a, b],
%  [{test, {check_pid, a}},
%   {test, {check_pid, b}},
%   {insert, result, {make_atom, true}}]}.



% %%% 9. Process dictionaries

erase(_) ->
    any().

% {def, put, [key, value],
%  [{bind, dict, {pid_dict, self}},
%   {insert, dict, {cons, key, value}},
%   {insert, result, {make_atom, undefined}},
%   {'or',
%    [{test, {eq, key, {car, dict}}},
%     {insert, result, {cdr, dict}}],
%    [{test, {not_atom, key}},
%     {test, {not_atom, {car, dict}}},
%     {insert, result, {cdr, dict}}]}]}.

put(any(), any()) ->
   special_definition().

% {def, get, [key],
%  [{bind, dict, {check_cons, {pid_dict, self}}},
%   {insert, result, {make_atom, undefined}},
%   {'or',
%    [{test, {eq, key, {car, dict}}},
%     {insert, result, {cdr, dict}}],
%    [{test, {not_atom, key}},
%     {test, {not_atom, {car, dict}}},
%     {insert, result, {cdr, dict}}]}]}.

get(_) ->
    special_definition().

get() ->
    any().
%    [any(), any()].
    

% %%% 10. Nodes

% {def, statistics, [type],
%  [{test, {eq, type, {make_atom, runtime}}},
%   {insert, result, {tuple, [{any_int}, {any_int}]}}]}.

% {def, node, [],
%  [{insert, result, {make_atom_any}}]}.

% {def, node, [something],
%  [{insert, result, {make_atom_any}}]}.

% %%% 11. Registries

register(atom(), pid()) ->
    true.

statistics(garbage_collection) ->
    {int(), int(), int()};
statistics(runtime) ->
    {int(), int()}.

whereis(atom()) ->
    pid().

% {def, register, [name, process],
%  [{test, {check_pid, process}},
%   {bind, reg, {node_registry, {pid_node, self}}},
%   {insert, reg, {cons, {check_atom, name}, process}},
%   {insert, result, {make_atom, true}}]}.

% {def, registered, [],
%  [{bind, res, {tag, registered_result, [pos]}},
%   {insert, res, {nil}},
%   {bind, name, {car, {node_registry, {pid_node, self}}}},
%   {insert, res, {cons, name, res}},
%   {insert, result, res}]}.

% {def, whereis, [name],
%  {'or',
%   [{bind, reg, {check_cons, {node_registry, {pid_node, self}}}},
%    {test, {eq, name, {car, reg}}},
%    {insert, result, {cdr, reg}}],
%   [{insert, result, {make_atom, undefined}}]}}.


% %%% 13. Miscellaneous

% {def, now, [],
%  {insert, result, {tuple, [{any_int}, {any_int}, {any_int}]}}}.

% {def, time, [],
%  {insert, result, {tuple, [{any_int}, {any_int}, {any_int}]}}}.

make_ref() ->
    ref().

now() ->
    {int(), int(), int()}.

localtime() ->
    {{int(), int(), int()},{int(), int(), int()}}.
    
date() ->
    {int(), int(), int()}.

phash(any(), int()) ->
    int().

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Operators %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% (will move all operator here eventually)

%%% 3b. Arithmetic operators

'+'(int(), int()) ->
    int();
'+'(int(), float()) ->
    float();
'+'(float(), int()) ->
    float();
'+'(float(), float()) ->
    float().

'-'(int(), int()) ->
    int();
'-'(int(), float()) ->
    float();
'-'(float(), int()) ->
    float();
'-'(float(), float()) ->
    float().

'-'(int()) ->
    int();
'-'(float()) ->
    float().

'*'(int(), int()) ->
    int();
'*'(int(), float()) ->
    float();
'*'(float(), int()) ->
    float();
'*'(float(), float()) ->
    float().

'/'(number(), number()) ->
    float().

'bor'(int(), int()) ->
    int().

'bxor'(int(), int()) ->
    int().

'bsl'(int(), int()) ->
    int().

'bsr'(int(), int()) ->
    int().

'band'(int(), int()) ->
    int().

'bnot'(int()) ->
    int().

'<'(_, _) ->
    bool().

'>'(_, _) ->
    bool().

'=<'(_, _) ->
    bool().

'>='(_, _) ->
    bool().

'=='(_, _) ->
    bool().

'=/='(_, _) ->
    bool().

'/='(_, _) ->
    bool().

'=:='(any(), any()) ->
    bool().

'++'(list(X), list(X)) ->
    list(X).

append(list(X), list(X)) ->
    list(X).

'--'(list(X), list(X)) ->
    list(X).

subtract(list(X), list(X)) ->
    list(X).

% {def, '>', [arg1, arg2],
%  [{insert, result, {make_atom, true}},
%   {insert, result, {make_atom, false}}]}.

% {def, '==', [arg1, arg2],
%  [{insert, result, {make_atom, true}},
%   {insert, result, {make_atom, false}}]}.

% {def, '++', [arg1, arg2],
%   {apply, {make_atom, '*internal*'}, {make_atom, '++'}, {cons, arg1, {cons, arg2, {nil}}},
%    result, self, except}}.

% {def, '--', [arg1, arg2],
%   {apply, {make_atom, '*internal*'}, {make_atom, '--'}, {cons, arg1, {cons, arg2, {nil}}},
%    result, self, except}}.

%%% Logical operators

'and'(bool(), bool()) -> 
    bool().

'or'(bool(), bool()) -> 
    bool().

'not'(bool()) -> 
    bool().


% foo() ->
%     special_definition().
% Used for experiments & debugging.

%%% Binaries

binary_to_list(binary()) ->
    list(int()).

binary_to_list(binary(), int(), int()) ->
    int().

list_to_binary(list(int())) ->
    binary().


%%%

get_module_info(any()) -> any().
get_module_info(any(),any()) -> any().
