-module(lists).

-type list(X) = [] + [X | list(X)].
-type deeplist(X) = [] + [deeplist(X) | deeplist(X)] + X.

-type bool() = true + false.
-type number() = int() + float().
-type string() = list(int()).

append(list(A),list(A)) -> list(A).

append(list(list(A))) -> list(A).

subtract(list(A), list(A)) -> list(A).

reverse(list(A)) -> list(A).
reverse(list(A), list(A)) -> list(A).

nth(int(), list(A)) -> A.
nthtail(int(), list(A)) -> list(A).

%% The list after dropping <tt>n</tt> elements

prefix(list(A),list(B)) -> bool().
suffix(list(A),list(B)) -> bool().
last(list(A)) -> A.

member(A,list(A)) -> bool().
seq(int(), int()) -> list(int()).
seq(int(), int(), int()) -> list(int()).

sum(list(int())) -> int().
sum(list(int()),int()) -> int().

duplicate(int(), A) -> list(A).

%% returns a list of the specified number of copies (>=0) 
%% of the second argument.

min(list(A)) -> A.
max(list(A)) -> A.

sublist(list(A), int(), int()) -> list(A).
sublist(list(A), int()) -> list(A).


%+type delete(A, [A]) -> [A].

delete(X, list(X)) ->
    list(X).

%% Deletes the <b>first</b> occurrence of the indicated element.

sort(list(X)) ->
    list(X).

% split_and_sort(list(X), list(X), list(X)) ->
%     list(X).

merge(list(list(X))) ->
    list(X).	

merge(list(X),list(X)) ->
    list(X).

merge(list(X),list(X),list(X)) ->
    list(X).

%+type sort([A]) -> [A].
%+type split_and_sort([A],[A],[A]) -> [A].
%+type merge([A], [A]) -> [A].
%+type merge([A], [A],[A]) -> [A].


thing_to_list(int() + float() + atom() + string()) ->
    string().

concat(list(int() + float() + atom() + string())) -> string().

%+type thing_to_list(integer() + float() + atom() + string()) -> 
%        string().
%+type concat([integer() + float() + atom() + string()]) -> string().


%+type flatten(deeplist(term())) -> [term()].
%+type flatten(deeplist(term()),[term()]) -> [term()].
%+type flatlength(deeplist(term())) -> integer().
%+type flatlength(deeplist(term()),integer()) -> integer().

flatten(deeplist(X)) ->
	       list(X).

flatten(deeplist(any()),list(any())) ->
	       list(any()).

flatlength(deeplist(any())) ->
    int().

flatlength(deeplist(any()), int()) ->
    int().


%+type keymember(Key, index :: integer(), [tuple()]) -> 
%        bool().
%+type keysearch(Key, index :: integer(), [tuple()]) ->
%        false | {value,tuple()}.
%+type keydelete(Key, index :: integer(), [tuple()]) -> 
%        [tuple()].
%+type keyreplace(Key, index :: integer(), [tuple()], tuple()) ->
%        [tuple()].
%+type keysort(index :: integer(), [tuple()]) -> 
%        [tuple()].
%+type keymerge(index :: integer(), [tuple()], [tuple()]) ->
%        [tuple()].

keymember(K, int(), list(any())) ->
    bool().
keysearch(K, int(), list(tuple(T))) ->
    false + {value, T}.
keydelete(K, int(), list(T)) ->
    list(T).
keyreplace(K, int(), list(T), T) ->
    list(T).
keysort(int(), list(T)) ->
    list(T).
keymerge(int(), list(T), list(T)) ->
    list(T).


%% Both lists have to be sorted before calling this function.

%+type keymap(fun(A)->B, index :: integer(), [tuple()]) -> 
%        [tuple()].
%+type keymap(fun(A)->B, [C], index :: integer(), [tuple()]) ->
%        [tuple()].
% ??? How should keymap/4be typed?? Note that the above type is not correct!!

keymap(fun(A) -> B end, int(), list(tuple(L))) ->
    list(tuple(L)).
	       

%+type split_and_keysort([tuple()], [tuple()], [tuple()], index :: integer()) ->
%        [tuple()].

% split_and_keysort(list(T), list(T), list(T), int()) ->
%     list(T).
      
%+type all(fun(A)->bool(), [A]) -> bool().

all(fun(A)->bool() end, list(A)) -> bool().
any(fun(A)->bool() end, list(A)) -> bool().

%+type takewhile(fun(A)->bool(), [A]) -> [A].
%+type dropwhile(fun(A)->bool(), [A]) -> [A].
%+type splitwith(fun(A)->bool(), [A]) -> {[A],[A]}.

takewhile(fun(A) -> bool() end, list(A)) -> list(A).
dropwhile(fun(A) -> bool() end, list(A)) -> list(A).
splitwith(fun(A) -> bool() end, list(A)) -> {list(A), list(A)}.

filter(fun(A)->bool() end, list(A)) -> list(A).

map(fun(A) -> B end, list(A)) -> list(B).
%+type flatmap(fun(A)->[B], [A]) -> [B].

% According to the OTP R9/B documentation, the type of flatmap should
% be
%
% flatmap(fun (A) -> B end, list(A)) -> list(B).
%
% However, I have obtained better results with the following type.

flatmap(fun (A) -> list(B) end, list(A)) -> list(B).

foldl(fun(A,B)->B end, B, list(A)) -> B.

%+type foldr(fun(A,B)->B, B, [A]) -> B.

foldr(fun(A,B)->B end, B, list(A)) -> B.

%+type mapfoldl(fun(A,B)->{D,B}, B, [A]) -> {[D],B}.
%+type mapfoldr(fun(A,B)->{D,B}, B, [A]) -> {[D],B}.

mapfoldl(fun(A, B) -> {D, B} end, B, list(A)) ->
    {list(D), B}.

mapfoldr(fun(A, B) -> {D, B} end, B, list(A)) ->
    {list(D), B}.

%+type foreach(fun(A)->B, [A]) -> ok.

foreach(fun(A) -> B end, list(A)) -> ok.		 

%%
%% the following functions became obsolete by the introduction
%%  of the <b>Funs</b> data structure in <tt>Erlang</tt> 4.4.
%%
%+type map(fun(A) -> B, [C], [D]) -> obsolete().
%+type all(fun(A)->bool(), [C], [D]) -> obsolete().
%+type any(fun(A)->bool(), [C], [D]) -> obsolete().
%+filter(fun(A)->bool(), [C], [D]) -> obsolete().
%+type flatmap(fun(A)->[B], [C], [D]) -> obsolete().
%+type foldl(fun(A)->B, [C], B, [D]) -> obsolete().
%+type foldr(fun(A)->B, [C], B, [D]) -> obsolete().
%+type foreach(fun(A)->B, [C], [D]) -> obsolete().
%+type mapfoldl(fun(A,B)->{D,B},[C], B, [E]) -> obsolete().
%+type mapfoldr(fun(A,B)->{D,B},[C], B, [E]) -> obsolete().


%+private.

%+usedeftype([number/0,string/0,int/0,deeplist/1]).

